"""Misc functions."""
import abc
import asyncio
import hashlib
import secrets
import string
from typing import Any, Awaitable, Callable, Iterable

import lorem_pysum
from jsonrpcobjects.objects import ParamsRequest, Request, ResultResponse
from openrpc import Example, Method, ParamStructure, Schema, SchemaType

_Security = dict[str, list[str]] | None
_Depends = dict[str, Any] | None
Middleware = Callable[
    [dict[str, str]], tuple[_Depends, _Security] | Awaitable[tuple[_Depends, _Security]]
]
AsyncMiddleware = Callable[[dict[str, str]], Awaitable[tuple[_Depends, _Security]]]
RequestProcessor = Callable[[str, dict[str, Any]], Awaitable[str | None]]


class OAuthChallengeHandler(abc.ABC):
    """Determine behavior for creating and getting OAuth verifiers."""

    @abc.abstractmethod
    def create_verifier(self) -> str:
        """Create a challenge to be used by OAuth and return the hash."""

    @abc.abstractmethod
    def get_verifier(self, challenge: str) -> str | None:
        """Get a verifier from a challenge."""


class DefaultHandler(OAuthChallengeHandler):
    def __init__(self) -> None:
        self._challenges: dict[str, str] = {}
        self._challenge_length = 128

    def get_verifier(self, challenge: str) -> str | None:
        return self._challenges.get(challenge)

    def create_verifier(self) -> str:
        characters = string.ascii_letters + string.digits
        challenge = "".join(
            secrets.choice(characters) for _ in range(self._challenge_length)
        )
        sha256_hash = hashlib.sha256(challenge.encode()).hexdigest()
        self._challenges[sha256_hash] = challenge
        return sha256_hash


def get_schema_from_input_ids(
    schema: SchemaType, input_ids: Iterable[int]
) -> SchemaType:
    """Get the appropriate schema from input ids."""
    for id_ in input_ids:
        schema = _get_schema_from_input_id(schema, id_)
    return schema


def _get_schema_from_input_id(schema: SchemaType, input_id: int) -> SchemaType:
    """Get the appropriate schema from an input id."""
    if isinstance(schema, bool):
        return schema
    if schema.items:
        return schema.items
    if schema.additional_properties:
        return schema.additional_properties
    if schema.any_of:
        return schema.any_of[input_id]
    return schema


def get_examples(methods: list[Method]) -> dict[str, tuple[str, str]]:
    """Get request/response examples for each method with examples.

    The keys in the returned dictionary will be method names, the values
    will be a tuple of (request example, response example).
    """
    examples = {}
    for method in methods:
        if not method.examples:
            continue
        param_example = method.examples[0].params
        result_example = method.examples[0].result
        if param_example is None or result_example is None:
            continue
        examples[method.name] = (
            _get_example_request(method.name, param_example, method.param_structure),
            _get_example_response(result_example),
        )
    return examples


def _get_example_request(
    method: str, examples: list[Example], param_structure: ParamStructure | None
) -> str:
    """Get an example request."""
    if param_structure is ParamStructure.BY_POSITION or not all(
        e.name for e in examples
    ):
        example_params: list | dict = [e.value for e in examples]
    else:
        example_params = {e.name: e.value for e in examples}
    if example_params:
        request: Request | ParamsRequest = lorem_pysum.generate(
            ParamsRequest, overrides={"method": method, "params": example_params}
        )
    else:
        request = lorem_pysum.generate(Request, overrides={"method": method})
    return request.model_dump_json(indent=2, by_alias=True)


def _get_example_response(example: Example) -> str:
    """Get an example request."""
    return lorem_pysum.generate(
        ResultResponse, overrides={"result": example.value}
    ).model_dump_json(indent=2, by_alias=True)


def is_any(schema: SchemaType | None) -> bool:
    """Return true if the given schema is any."""
    if schema is None or schema is True:
        return True
    return schema is False or not any(
        (
            "type" in schema.model_fields_set,
            "ref" in schema.model_fields_set,
            "all_of" in schema.model_fields_set,
            "any_of" in schema.model_fields_set,
            "one_of" in schema.model_fields_set,
            "const" in schema.model_fields_set,
            "enum" in schema.model_fields_set,
            "not_" in schema.model_fields_set,
        )
    )


def get_any_default(schema: Schema) -> int:
    """Get default option of an `anyOf` schema."""
    default = 0
    if isinstance(schema, bool):
        return default
    for i, any_schema in enumerate(schema.any_of or []):
        if isinstance(any_schema, bool):
            continue
        if "default" in schema.model_fields_set and (
            (
                "const" in any_schema.model_fields_set
                and schema.default == any_schema.const
            )
            or (schema.default is None and any_schema.type == "null")
        ):
            default = i
        if any_schema.ref and i == default:
            default += 1
    return default


def get_middleware(middleware: Middleware | None) -> AsyncMiddleware | None:
    """Get middleware function as async function."""
    if middleware is None:
        return None

    if asyncio.iscoroutinefunction(middleware):

        async def _async_middleware(
            headers: dict[str, str]
        ) -> tuple[_Depends, _Security]:
            return await middleware(headers)

        return _async_middleware

    async def _middleware(headers: dict[str, str]) -> tuple[_Depends, _Security]:
        # Type ignore here because mypy fails to understand
        # `iscoroutinefunction` check above.
        return middleware(headers)  # type: ignore

    return _middleware



"""Provides RPCInterface to communicate with RPC server."""
import os
from pathlib import Path
from typing import Optional

from openrpc import Server

from tabella._templating_engine import TemplatingEngine
from tabella._util import DefaultHandler, OAuthChallengeHandler, RequestProcessor


class TabellaCache:
    """Handle communication with RPC server."""

    _instance: Optional["TabellaCache"] = None

    def __init__(self) -> None:
        if TabellaCache._instance:
            msg = "Do not instantiate RPCInterface directly, call `get_instance()`"
            raise ValueError(msg)
        self.api_url: str | None = os.environ.get("API_URL")
        self.cache: dict[str | None, TemplatingEngine] = {}
        self.request_processor: RequestProcessor | None = None
        self.servers: list[Server] = []
        self.oauth_handler: OAuthChallengeHandler = DefaultHandler()
        self.client_gen_directory: Path | None = None
        self.disable_client_gen: bool = False

    @staticmethod
    def get_instance() -> "TabellaCache":
        """Get or create instance of RPCInterface."""
        if TabellaCache._instance:
            return TabellaCache._instance
        TabellaCache._instance = TabellaCache()
        return TabellaCache._instance

    def set_request_processor(
        self, request_processor: RequestProcessor, url: str
    ) -> None:
        """Set RPCServer to use for discovery and request processing."""
        self.request_processor = request_processor
        self.api_url = url

    async def get_templating_engine(
        self, api_url: str | None = None, *, refresh: bool = False
    ) -> TemplatingEngine:
        """Get a templating engine for the given API URL."""
        api_url = api_url or self.api_url
        if (te := self.cache.get(api_url)) and not refresh:
            return te
        if self.api_url and self.request_processor and not refresh:
            self.cache[api_url] = TemplatingEngine(self.api_url, self.request_processor)
        elif api_url is None:
            self.cache[api_url] = TemplatingEngine(None)
        else:
            self.cache[api_url] = TemplatingEngine(api_url, self.request_processor)
        await self.cache[api_url].init()
        return self.cache[api_url]



"""Used to handle templating variables and string sanitization."""
import importlib
import inspect
import json
from copy import copy
from typing import Any

from jinja2 import Undefined
from openrpc import Contact, Method, OAuth2, OpenRPC, RPCServer, SchemaType, Tag

from tabella._util import RequestProcessor

try:
    websockets = importlib.import_module("websockets")
except ImportError:
    websockets = None  # type: ignore
try:
    httpx = importlib.import_module("httpx")
except ImportError:
    httpx = None  # type: ignore

_rpc_server = RPCServer(
    title="Tabella",
    version="1.0.0",
    description="OpenRPC Interactive API Documentation",
    contact=Contact(
        name="Tabella Source Code",
        url="https://gitlab.com/mburkard/tabella",
    ),
)


class TemplatingEngine:
    """Get templating variables."""

    def __init__(
        self, api_url: str | None, request_processor: RequestProcessor | None = None
    ) -> None:
        self.tags: list[Tag] = []
        self.api_url = api_url
        self.request_processor = request_processor
        self.rpc = OpenRPC(**_rpc_server.discover())
        self.headers: dict[str, str] = {}

    async def init(self) -> None:
        """Init a Templating Engine."""
        if self.api_url is None:
            self.rpc = await self._get_rpc(OpenRPC(**_rpc_server.discover()))
        else:
            self.rpc = await self._get_rpc()
        # Get all tags used in methods.
        for method in self.rpc.methods:
            for tag in method.tags or []:
                if tag not in self.tags:
                    self.tags.append(tag)

    def get_tags(self, method: Method) -> list[int]:
        """Gat indexes of tags for the given method."""
        return [
            i for i, tag in enumerate(self.tags or []) if tag in (method.tags or [])
        ]

    def get_method(self, idx: int) -> Method:
        """Gat method with the given index."""
        for i, method in enumerate(self.rpc.methods):
            if i == idx:
                return method
        msg = f"No method at index {idx}."
        raise ValueError(msg)

    async def _get_rpc(self, rpc_doc: OpenRPC | None = None) -> OpenRPC:
        if not rpc_doc:
            discover = {"id": 1, "method": "rpc.discover", "jsonrpc": "2.0"}
            resp = json.loads(await self.process_request(discover))
            rpc_doc = OpenRPC(**resp["result"])
        if rpc_doc.components is None or rpc_doc.components.schemas is None:
            return rpc_doc

        # Resolve references in schemas and filter methods by server.
        methods = []
        for method in rpc_doc.methods:
            if method.servers and self.api_url not in [s.url for s in method.servers]:
                continue
            for param in method.params:
                param.schema_ = resolve_references(
                    param.schema_, rpc_doc.components.schemas
                )
            method.result.schema_ = resolve_references(
                method.result.schema_, rpc_doc.components.schemas
            )
            methods.append(method)
        rpc_doc.methods = methods
        return rpc_doc

    async def process_request(self, request: dict[str, Any]) -> str:
        """Send a request to the RPC API and get a response."""
        # Get method from methodId.
        if self.rpc:
            for i, m in enumerate(self.rpc.methods):
                if str(i) == request["method"]:
                    request["method"] = m.name
                    break

        rpc_request = json.dumps(request)

        # Custom Request Processor.
        if self.request_processor is not None:
            resp = self.request_processor(rpc_request, self.headers)
            return await resp if inspect.isawaitable(resp) else resp

        # HTTP
        headers = {**{"Content-Type": "application/json"}, **self.headers}
        if self.api_url and self.api_url.startswith("http"):
            if httpx is None:
                msg = "httpx must be installed for HTTP APIs."
                raise ImportError(msg)
            client = httpx.AsyncClient()
            return (
                await client.post(
                    self.api_url,
                    content=rpc_request,
                    headers=headers,
                )
            ).content.decode()

        # Websocket
        if self.api_url and self.api_url.startswith("ws"):
            if websockets is None:
                msg = "websockets must be installed for WebSocket APIs."
                raise ImportError(msg)
            # Type ignore because `websockets` typing is wrong.
            async with websockets.connect(  # type: ignore
                self.api_url, extra_headers=headers
            ) as websocket:
                await websocket.send(rpc_request)
                return await websocket.recv()
        msg = "Invalid API URL."
        raise ValueError(msg)

    def get_security_id(self, security: dict[str, Any]) -> str:
        """Get ID to use for a security icon."""
        if (
            self.rpc.components is None
            or self.rpc.components.x_security_schemes is None
        ):
            return ""
        id_ = ""
        # Find matching scheme name.
        for s_name in security:
            for i, name in enumerate(self.rpc.components.x_security_schemes):
                if name != s_name:
                    continue
                id_ += f"scheme_{i}"
                scheme = self.rpc.components.x_security_schemes[name]
                # Find scope indexes.
                if not isinstance(scheme, OAuth2):
                    continue
                for j, scope in enumerate(scheme.flows[0].scopes):
                    for s_scope in security[name]:
                        if scope == s_scope:
                            id_ += f"_{j}"
        return id_

    def is_locked(
        self, method_id: str, enabled_scheme: str | Undefined, enabled_scopes: list[str]
    ) -> bool:
        """Determine if security method is locked."""
        if (
            self.rpc.components is None
            or self.rpc.components.x_security_schemes is None
            or enabled_scheme == Undefined
        ):
            return True

        method = self.get_method(int(method_id))
        if not method.x_security:
            return True
        for i, name in enumerate(self.rpc.components.x_security_schemes):
            if str(i) == enabled_scheme:
                if not (required_scopes := method.x_security.get(name)):
                    return True
                return not all(scope in enabled_scopes for scope in required_scopes)

        return True


def resolve_references(
    schema: SchemaType,
    schemas: dict[str, SchemaType],
    recurring: list[SchemaType] | None = None,
) -> SchemaType:
    """Resolve JSON Schema references."""
    if isinstance(schema, bool):
        return schema

    # Don't mutate original schema, will cause infinite recursion on
    # subsequent `resolve_references` calls with the same root schema.
    schema = copy(schema)

    # Don't mutate original recursion list, if mutated sibling schemas
    # will show as recursive for all but the first one.
    recurring = copy(recurring) if recurring else []

    if schema.ref:
        ref = schema.ref.removeprefix("#/components/schemas/")
        resolved_ref = copy(schemas[ref])
        if isinstance(resolved_ref, bool):
            return resolved_ref
        if schema in recurring:
            # Set `ref` to indicate recursion.
            resolved_ref.ref = schema.ref
            return resolved_ref
        resolved_ref.ref = None
        recurring.append(schema)
        schema = resolved_ref

    # Lists of schemas.
    for attr in ["all_of", "any_of", "one_of", "prefix_items"]:
        resolved = []
        for child_schema in getattr(schema, attr) or []:
            resolved_option = resolve_references(child_schema, schemas, recurring)
            resolved.append(resolved_option)
        if resolved:
            setattr(schema, attr, resolved)

    # Single schemas.
    for attr in [
        "not_",
        "property_names",
        "items",
        "contains",
        "if_",
        "then",
        "else_",
        "additional_properties",
    ]:
        if getattr(schema, attr):
            setattr(
                schema,
                attr,
                resolve_references(getattr(schema, attr), schemas, recurring),
            )

    # Dict of schemas.
    for attr in ["properties", "pattern_properties", "defs", "dependent_schemas"]:
        resolved_dict = {}
        for name, child_schema in (getattr(schema, attr) or {}).items():
            resolved_dict[name] = resolve_references(child_schema, schemas, recurring)
        if resolved_dict:
            setattr(schema, attr, resolved_dict)

    return schema



"""Tabella API function definitions."""
__all__ = ("get_app",)

import asyncio
import importlib
import json
import shutil
import sys
from pathlib import Path
from typing import Any
from urllib import parse

import caseswitcher
from fastapi import FastAPI, Request
from fastapi.responses import HTMLResponse
from fastapi.staticfiles import StaticFiles
from fastapi.templating import Jinja2Templates
from jsonrpcobjects.objects import (
    ErrorResponse,
    Notification,
    ParamsNotification,
    ParamsRequest,
    Request as RPCRequest,
    ResultResponse,
)
from openrpc import (
    OAuth2,
    OpenRPC,
    ParamStructure,
    RPCServer,
    Schema,
    SchemaType,
    Server,
)
from openrpcclientgenerator import generate, Language
from starlette.responses import FileResponse, JSONResponse, Response

# noinspection PyProtectedMember
from starlette.templating import _TemplateResponse
from starlette.websockets import WebSocket, WebSocketDisconnect

from tabella import _util as util
from tabella._cache import TabellaCache
from tabella._templating_engine import TemplatingEngine
from tabella._util import Middleware, OAuthChallengeHandler, RequestProcessor

try:
    httpx = importlib.import_module("httpx")
except ImportError:
    httpx = None  # type: ignore

RequestType = ParamsRequest | RPCRequest | ParamsNotification | Notification

root = Path(__file__).parent

app = FastAPI(docs_url="/swagger")
app.mount("/static", StaticFiles(directory=root / "static"), name="static")
templates = Jinja2Templates(
    directory=root / "templates", lstrip_blocks=True, trim_blocks=True
)

cache = TabellaCache.get_instance()
base_context: dict[str, Any] = {
    "len": len,
    "str": str,
    "id": lambda x: f"_{x}",
    "array_item_id": lambda x: f"_item{x}",
    "is_any": util.is_any,
    "is_str": lambda x: isinstance(x, str),
    "is_": lambda x, y: x is y,
    "ParamStructure": ParamStructure,
    "key_schema": Schema(type="string"),
    "httpx_missing": httpx is None,
}


def get_app(  # noqa: PLR0913
    rpc: RPCServer | None,
    middleware: Middleware | None = None,
    oauth_challenge_handler: OAuthChallengeHandler | None = None,
    client_gen_directory: Path | None = None,
    favicon_url: str | None = None,
    icon_url: str | None = None,
    *,
    disable_client_gen: bool = False,
) -> FastAPI:
    """Host the Given RPCServer.

    :param rpc: RPC server to host.
    :param middleware: Function to get RPCServer `depends` and
        `security` values from request headers.
    :param oauth_challenge_handler: Class to handle creating and getting
        OAuth verifiers from code challenges.
    :param client_gen_directory: Directory to write generated client
        files to.
    :param favicon_url: URL to favicon to use for docs site.
    :param icon_url: URL to icon to use in docs site.
    :param disable_client_gen: Disable client generation and download.
    """
    cache.disable_client_gen = disable_client_gen
    if not rpc:
        return app

    async_middleware = util.get_middleware(middleware)

    # Determine request processor method.
    if async_middleware is None:
        # Ignore and noqa because signatures must be identical.
        # noinspection PyUnusedLocal
        async def _request_processor(
            request: str, headers: dict  # noqa: ARG001
        ) -> str | None:
            return await rpc.process_request_async(request) or ""

    else:

        async def _request_processor(request: str, headers: dict) -> str | None:
            return (
                await rpc.process_request_async(
                    request, *await async_middleware(headers)
                )
                or ""
            )

    servers = _get_servers(rpc)
    for server in servers:
        parsed_url = parse.urlparse(server.url)
        if parsed_url.scheme.startswith("http"):
            _add_ws_api_url(parsed_url.path, _request_processor)
        elif parsed_url.scheme.startswith("ws"):
            _add_http_api_url(parsed_url.path, _request_processor)

    api_path = servers[0].url
    cache.set_request_processor(_request_processor, api_path)
    if client_gen_directory:
        cache.client_gen_directory = client_gen_directory
    if favicon_url:
        base_context["favicon_url"] = favicon_url
    if icon_url:
        base_context["icon_url"] = icon_url
    if oauth_challenge_handler:
        cache.oauth_handler = oauth_challenge_handler
    if len(servers) > 1:
        cache.servers = servers

    return app


def _get_servers(rpc: RPCServer) -> list[Server]:
    servers = []
    if not isinstance(rpc.servers, Server):
        servers = rpc.servers
    else:
        parsed_url = parse.urlparse(rpc.servers.url)
        if not (not parsed_url.scheme and parsed_url.path == "localhost"):
            servers = [rpc.servers]
        else:
            port = "8000"
            for arg in sys.argv:
                if arg == "--port":
                    port = sys.argv[sys.argv.index(arg) + 1]
                    break
            servers.append(Server(name="HTTP API", url=f"http://localhost:{port}/api"))
            servers.append(
                Server(name="WebSocket API", url=f"ws://localhost:{port}/api")
            )
    return servers


def _add_ws_api_url(api_path: str, request_processor: RequestProcessor) -> None:
    @app.websocket(api_path)
    async def ws_process_rpc(websocket: WebSocket) -> None:
        """Process RPC requests through websocket."""
        try:
            await websocket.accept()

            async def _process_rpc(request: str) -> None:
                rpc_response = await request_processor(request, {**websocket.headers})
                if rpc_response is not None:
                    await websocket.send_text(rpc_response)

            while True:
                data = await websocket.receive_text()
                asyncio.create_task(_process_rpc(data))
        except WebSocketDisconnect:
            pass


def _add_http_api_url(api_path: str, request_processor: RequestProcessor) -> None:
    @app.post(api_path, response_model=ErrorResponse | ResultResponse | None)
    async def http_process_rpc(request: Request, rpc_request: RequestType) -> Response:
        """Process RPC request through HTTP server."""
        rpc_response = await request_processor(
            rpc_request.model_dump_json(), {**request.headers}
        )
        return Response(content=rpc_response, media_type="application/json")


@app.get("/", response_class=HTMLResponse)
async def index(request: Request) -> _TemplateResponse:
    """Get interactive docs site."""
    scheme = None
    enabled_scopes = []
    client_id = None

    # If state is provided this is redirect from OAuth server.
    state = request.query_params.get("state")
    if not state:
        te = await cache.get_templating_engine()
    else:
        code = request.query_params.get("code")
        hash_, scheme, scopes, client_id, api_url = state.split("@")
        enabled_scopes = scopes.split(" ")
        if not (verifier := cache.oauth_handler.get_verifier(hash_)):
            return _error(request, "Server Error", "OAuth 2.0 verifier is missing.")

        te = await cache.get_templating_engine(api_url)
        # This shouldn't happen but needed for type safety.
        if te.rpc.components is None:
            return _error(
                request,
                "Server Error",
                "This server does not have an OAuth 2.0. security scheme",
            )

        schemes = te.rpc.components.x_security_schemes or {}
        for i, name in enumerate(schemes):
            if i != int(scheme):
                continue
            security_scheme = schemes[name]
            if not isinstance(security_scheme, OAuth2):
                continue
            token_url = security_scheme.flows[0].token_url
            client = httpx.AsyncClient()
            result = await client.post(
                f"{token_url}?code={code}&code_verifier={verifier}&grant_type=token"
            )
            te.headers = {"Authorization": f"Bearer {result.json()['access_token']}"}
            break

    context = {
        "request": request,
        "disable_api_input": cache.api_url is not None,
        "disable_client_gen": cache.disable_client_gen,
        "api_url": te.api_url,
        "te": te,
        "examples": util.get_examples(te.rpc.methods),
        "servers": cache.servers,
        "code_challenge": cache.oauth_handler.create_verifier(),
        "enabled_scheme": scheme,
        "enabled_scopes": enabled_scopes,
        "client_id": client_id,
    }
    return templates.TemplateResponse("index.html", {**context, **base_context})


@app.get("/docs/discover-{trigger}", response_class=HTMLResponse)
async def discover(request: Request) -> _TemplateResponse:
    """Get OpenRPC docs."""
    api_url = request.query_params.get("api-url")
    trigger = request.path_params["trigger"]
    te = await cache.get_templating_engine(api_url, refresh=trigger == "click")

    context = {
        "disable_client_gen": cache.disable_client_gen,
        "request": request,
        "te": te,
        "examples": util.get_examples(te.rpc.methods),
        "code_challenge": cache.oauth_handler.create_verifier(),
    }
    return templates.TemplateResponse("openrpc_docs.html", {**context, **base_context})


@app.get("/docs/try-it-modal/{method_idx}", response_class=HTMLResponse)
async def try_it(request: Request) -> _TemplateResponse:
    """Get "Try it out" modal for a method."""
    api_url = request.query_params.get("api-url")
    method_idx = request.path_params["method_idx"]
    te = await cache.get_templating_engine(api_url)
    method = te.get_method(int(method_idx))
    context = {
        "request": request,
        "method": method,
        "method_id": method_idx,
        "get_any_default": util.get_any_default,
    }
    return templates.TemplateResponse("modals/try_it.html", {**context, **base_context})


@app.get(
    "/docs/add-{item_type}-item/{method_id}/{param_id}/{input_id}/{is_even}",
    response_class=HTMLResponse,
)
async def add_item(request: Request) -> _TemplateResponse:
    """Get "Try it out" modal for a method."""
    api_url = request.query_params.get("api-url")
    item_count = request.query_params.get("item-count")
    method_id = request.path_params["method_id"]
    param_id = int(request.path_params["param_id"])
    input_id = request.path_params["input_id"]
    item_type = request.path_params["item_type"]
    is_even = int(request.path_params["is_even"])
    te = await cache.get_templating_engine(api_url)
    method = te.get_method(int(method_id))

    # Get schema for this param.
    schema: SchemaType = Schema()
    for i, param in enumerate(method.params):
        if i == param_id:
            schema = param.schema_
            break

    # We already got param schema from method, remove those ids.
    input_tree_path = input_id.removeprefix(f"{method_id}_{param_id}")
    # Get input ids to get proper schema in schema tree.
    input_ids = [id_.removeprefix("item") for id_ in input_tree_path.split("_") if id_]
    schema = util.get_schema_from_input_ids(schema, map(int, input_ids))

    if item_type == "array" and not isinstance(schema, bool) and schema.items:
        schema = schema.items
    if item_type != "recursive":
        input_id = f"{input_id}_item{item_count}"

    context = {
        "request": request,
        "method_id": method_id,
        "param_id": str(param_id),
        "schema": schema,
        "input_id": input_id,
        "minimalize": True,
        "required": True,
        "get_any_default": util.get_any_default,
        "is_even": 0 if is_even else 1,
    }
    if item_type == "object":
        return templates.TemplateResponse(
            "schema_form/object.html", {**context, **base_context}
        )
    return templates.TemplateResponse(
        "schema_form/form.html", {**context, **base_context}
    )


@app.get("/openrpc.json")
async def openrpc_doc(request: Request) -> JSONResponse:
    """Get raw OpenRPC JSON document."""
    api_url = request.query_params.get("api-url")
    te = await cache.get_templating_engine(api_url)
    return JSONResponse(content=await _discover(te))


@app.post("/rpc-api")
async def api_pass_through(request: Request) -> Response:
    """Pass RPC requests to RPC server and get response."""
    api_url = request.query_params.get("api-url")
    te = await cache.get_templating_engine(api_url)

    # Set request headers based on security scheme.
    components = te.rpc.components
    scheme_is_set = False
    if components and components.x_security_schemes:
        for i, name in enumerate(components.x_security_schemes):
            if value := request.headers.get(f"scheme{i}"):
                scheme_is_set = True
                scheme = components.x_security_schemes[name]
                if scheme.type == "bearer":
                    te.headers = {scheme.name: f"Bearer {value}"}
                elif scheme.type == "apikey":
                    te.headers = {scheme.name: value}
                break
    # If no scheme is set clear any existing headers.
    if not scheme_is_set:
        te.headers = {}

    response = await te.process_request(await request.json())
    return Response(content=response, media_type="application/json")


@app.get("/download-client")
async def download_client(request: Request) -> FileResponse:
    """Download a generated client for this API."""
    if cache.disable_client_gen:
        msg = "Client generation is disabled."
        raise ValueError(msg)
    # Get RPC data and target language.
    api_url = request.query_params["api-url"]
    language = request.query_params["language"]
    te = await cache.get_templating_engine(api_url)
    rpc = OpenRPC(**await _discover(te))
    lang_option = Language.PYTHON if language == "Python" else Language.TYPESCRIPT

    # Make generated out directories if they don't exist.
    out = cache.client_gen_directory or root.joinpath("static/out")
    out.mkdir(exist_ok=True)
    lang_out = out.joinpath(language.lower())
    lang_out.mkdir(exist_ok=True)
    transport = "http" if api_url.startswith("http") else "ws"
    client_name = caseswitcher.to_kebab(rpc.info.title) + f"-{transport}-client"
    client_dir = lang_out.joinpath(client_name)
    filename = f"{client_name}-{rpc.info.version}-{lang_option.value}.zip"
    zip_file = lang_out.joinpath(filename)

    # If client doesn't exist, generate and zip it.
    if not zip_file.exists():
        generate(rpc, lang_option, api_url, out)
        shutil.make_archive(zip_file.as_posix().removesuffix(".zip"), "zip", client_dir)

    # Serve the client zipfile.
    return FileResponse(zip_file, headers={"Content-Disposition": filename})


async def _discover(te: TemplatingEngine) -> dict[str, Any]:
    discover_request = {"id": 1, "method": "rpc.discover", "jsonrpc": "2.0"}
    response = await te.process_request(discover_request)
    return json.loads(response)["result"]


def _error(request: Request, error_title: str, error_message: str) -> _TemplateResponse:
    context = {
        "request": request,
        "error_title": error_title,
        "error_message": error_message,
    }
    return templates.TemplateResponse("error.html", context)



"""Documentation and hosting for OpenRPC APIs."""
__all__ = ("get_app", "Middleware", "OAuthChallengeHandler")

from tabella._app import get_app
from tabella._util import Middleware, OAuthChallengeHandler



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['tabella']

package_data = \
{'': ['*'],
 'tabella': ['static/*',
             'templates/*',
             'templates/modals/*',
             'templates/modals/auth/*',
             'templates/schema/*',
             'templates/schema_form/*']}

install_requires = \
['Jinja2>=3.1.2,<4.0.0',
 'case-switcher>=1.3.13,<2.0.0',
 'fastapi>=0.103.2,<0.104.0',
 'httpx>=0.25.0,<0.26.0',
 'jsonrpc2-objects>=3.0.0,<4.0.0',
 'lorem-pysum>=1.4.3,<2.0.0',
 'openrpc>=8.4.1,<9.0.0',
 'openrpcclientgenerator>=0.46.0,<0.47.0',
 'pydantic>=2.4.0,<3.0.0']

setup_kwargs = {
    'name': 'tabella',
    'version': '1.0.0',
    'description': 'Open-RPC API hosting and interactive documentation.',
    'long_description': '# Tabella\n\n![](https://img.shields.io/badge/License-ApacheV2-blue.svg)\n![](https://img.shields.io/badge/code%20style-black-000000.svg)\n![](https://img.shields.io/pypi/v/tabella.svg)\n\n## Open-RPC development framework with builtin interactive documentation.\n\n![Demo](https://gitlab.com/mburkard/tabella/-/raw/main/docs/demo.png)\n\n## Live Demo\n\nA live demo is available [here](https://tabella.burkard.cloud/).\n\n## Install\n\nTabella is on PyPI and can be installed with:\n\n```shell\npip install tabella\n```\n\nOr with [Poetry](https://python-poetry.org/)\n\n```shell\npoetry add tabella\n```\n\n## Python OpenRPC Docs\n\nThe RPC server hosted and documented by Tabella is powered\nby [Python OpenRPC](https://gitlab.com/mburkard/openrpc). Refer to the Python OpenRPC\ndocs hosted [here](https://python-openrpc.burkard.cloud/) for advanced use.\n\n## Getting Started\n\nA basic Tabella app:\n\n```python\nimport tabella\nfrom openrpc import RPCServer\nimport uvicorn\n\nrpc = RPCServer()\n\n\n@rpc.method()\ndef echo(a: str, b: float) -> tuple[str, float]:\n    """Echo parameters back in result."""\n    return a, b\n\n\napp = tabella.get_app(rpc)\n\nif __name__ == "__main__":\n    uvicorn.run(app, host="127.0.0.1", port=8000)\n```\n\nRun this, then open http://127.0.0.1:8000/ in your browser to use the interactive\ndocumentation.\n\nThe Open-RPC API will be hosted over HTTP on `http://127.0.0.1:8000/api` and over\nWebSockets on `ws://127.0.0.1:8000/api`.\n\n## Further Usage\n\n### Security and Depends Arguments\n\nThe `tabella.get_app` function accepts argument, `middleware`, which is a callable that\nwill be passed HTTP request headers/WebSocket connection headers. The `middleware`\nfunction must return a tuple of two dictionaries, the first will be passed to the\nRPC Server as\n[Depends Arguments](https://python-openrpc.burkard.cloud/security/depends_arguments)\nand the second as\n[Security Arguments](https://python-openrpc.burkard.cloud/security/schemes).\n\n### Set Servers\n\nSet RPC servers manually to specify transport and paths to host the RPC server on, e.g.\n\n```python\nfrom openrpc import RPCServer, Server\nimport tabella\n\nrpc = RPCServer(\n    servers=[\n        Server(name="HTTP API", url="http://localhost:8000/my/api/path"),\n        Server(name="WebSocket API", url="ws://localhost:8000/my/api/path"),\n    ]\n)\napp = tabella.get_app(rpc)\n```\n\nThis app will host the RPCServer over HTTP and over WebSockets with the\npath `/my/api/path`.\n\n### Pydantic\n\n[Pydantic](https://docs.pydantic.dev/latest/) is used for request/response\ndeserialization/serialization as well as schema generation. Pydantic should be used for\nany models as seen here in\nthe [Python OpenRPC Docs](https://python-openrpc.burkard.cloud/basics#pydantic-for-data-models).\n\n### FastAPI\n\nTabella HTTP and WebSocket server hosting is uses\n[FastAPI](https://fastapi.tiangolo.com/). The app returned by `tabella.get_app` is a\nFastAPI app.\n\n## Inspired By\n\n- [OPEN-RPC Playground](https://playground.open-rpc.org/)\n- [Swagger](https://swagger.io/)\n- [Redoc](https://github.com/Redocly/redoc)\n\n## Support The Developer\n\n<a href="https://www.buymeacoffee.com/mburkard" target="_blank">\n  <img src="https://cdn.buymeacoffee.com/buttons/v2/default-blue.png"\n       width="217"\n       height="60"\n       alt="Buy Me A Coffee">\n</a>\n',
    'author': 'Matthew Burkard',
    'author_email': 'matthewjburkard@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://gitlab.com/mburkard/tabella',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)
